package org.jboss.test.jbeap2263reproducer;

import com.google.common.base.Strings;
import org.apache.commons.io.IOUtils;
import org.apache.http.HttpEntity;
import org.apache.http.StatusLine;
import org.apache.http.client.CookieStore;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpUriRequest;
import org.apache.http.client.methods.RequestBuilder;
import org.apache.http.impl.client.BasicCookieStore;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.impl.client.HttpClients;
import org.apache.http.util.EntityUtils;
import org.junit.Assert;
import org.junit.BeforeClass;
import org.junit.Test;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.nio.charset.Charset;

/**
 * Note that if JBEAP-2263 is present, the test fails only once after the servers are started.
 * Subsequent runs pass, server restart is needed to get the test to fail again.
 */
public class FirstLogoutOnNonOwnerNodeTest {
    private static final Logger LOGGER = LoggerFactory.getLogger(FirstLogoutOnNonOwnerNodeTest.class);

    // the nodes should be on the same host - else the cookies for a single node won't matter to the other
    private static final String TEST_PROP_HOST = "testProp.host";
    private static final String TEST_PROP_PORT_1 = "testProp.port1";
    private static final String TEST_PROP_PORT_2 = "testProp.port2";
    private static final String HOST_DEFAULT = "127.0.0.1";
    private static final String PORT_1_DEFAULT = "8080";
    private static final String PORT_2_DEFAULT = "8180";

    private static final String AUTH_CHALLENGE_STRING = "j_security_check";
    private static final String CLUSTERBENCH_CONTEXT = "/clusterbench/session";
    private static final String CLUSTERBENCH_AUTH_CONTEXT = CLUSTERBENCH_CONTEXT.replace("session", AUTH_CHALLENGE_STRING);
    private static final String CLUSTERBENCH_INVALIDATE_OPTION = "?invalidate=true";

    private static final String USER = "user";
    private static final String PASSWORD = "secret_Passw0rd";

    private static String node1Address;
    private static String node2Address;

    @BeforeClass
    public static void setUpClass() {
        node1Address = "http://" + getPropertyOrDefault(TEST_PROP_HOST, HOST_DEFAULT) + ":" + getPropertyOrDefault(TEST_PROP_PORT_1, PORT_1_DEFAULT);
        node2Address = "http://" + getPropertyOrDefault(TEST_PROP_HOST, HOST_DEFAULT) + ":" + getPropertyOrDefault(TEST_PROP_PORT_2, PORT_2_DEFAULT);

        LOGGER.debug("Node 1: {}", node1Address);
        LOGGER.debug("Node 2: {}", node2Address);
    }

    @Test
    public void firstLogoutOnNonOwnerNode() throws IOException {
        CookieStore basicCookieStore = new BasicCookieStore();
        try (CloseableHttpClient closeableHttpClient = HttpClients.custom().setDefaultCookieStore(basicCookieStore).build()) {
            // first GET - verify we're not authenticated (node1)
            String body = getResponseBody(closeableHttpClient.execute(new HttpGet(node1Address + CLUSTERBENCH_CONTEXT)));
            logRequestResponse(node1Address + CLUSTERBENCH_CONTEXT, body);
            Assert.assertTrue("The first request does not require authentication", body.contains(AUTH_CHALLENGE_STRING));

            // first POST - authenticate (node1)
            HttpUriRequest httpUriRequest = RequestBuilder
                    .post()
                    .setUri(node1Address + CLUSTERBENCH_AUTH_CONTEXT)
                    .addParameter("j_username", USER)
                    .addParameter("j_password", PASSWORD)
                    .build();
            body = getResponseBody(closeableHttpClient.execute(httpUriRequest));
            logRequestResponse(node1Address + CLUSTERBENCH_AUTH_CONTEXT, body);
            Assert.assertFalse("First authentication did not succeed", body.contains(AUTH_CHALLENGE_STRING));

            // second GET - verify we're authenticated (node1)
            body = getResponseBody(closeableHttpClient.execute(new HttpGet(node1Address + CLUSTERBENCH_CONTEXT)));
            logRequestResponse(node1Address + CLUSTERBENCH_CONTEXT, body);
            Assert.assertFalse("First authentication did not succeed - subsequent GET failed", body.contains(AUTH_CHALLENGE_STRING));

            // third GET - verify we're authenticated on second node (node2)
            body = getResponseBody(closeableHttpClient.execute(new HttpGet(node2Address + CLUSTERBENCH_CONTEXT)));
            logRequestResponse(node2Address + CLUSTERBENCH_CONTEXT, body);
            Assert.assertFalse("First authentication did not succeed - we're not authenticated on second node", body.contains(AUTH_CHALLENGE_STRING));

            // fourth GET - log out from second node (node2)
            body = getResponseBody(closeableHttpClient.execute(new HttpGet(node2Address + CLUSTERBENCH_CONTEXT + CLUSTERBENCH_INVALIDATE_OPTION)));
            logRequestResponse(node2Address + CLUSTERBENCH_CONTEXT + CLUSTERBENCH_INVALIDATE_OPTION, body);

            // fifth GET - verify we need to authenticate (node2)
            body = getResponseBody(closeableHttpClient.execute(new HttpGet(node2Address + CLUSTERBENCH_CONTEXT)));
            logRequestResponse(node2Address + CLUSTERBENCH_CONTEXT, body);
            Assert.assertTrue("We're still authenticated after we logged out (invalidated the last session)", body.contains(AUTH_CHALLENGE_STRING));
        }
    }

    private static String getPropertyOrDefault(String propName, String defaultValue) {
        String propValue = System.getProperty(propName);
        return Strings.isNullOrEmpty(propValue) ? defaultValue : propValue;
    }

    private static void logRequestResponse(String address, String responseBody) {
        LOGGER.trace("Request to {}, response:\n{}", address, responseBody);
    }

    private static void logCookies(CookieStore cookieStore) {
        LOGGER.trace("Cookies present in the cookie store: {}", cookieStore.getCookies());
    }

    private String getResponseBody(CloseableHttpResponse closeableHttpResponse) throws IOException {
        StatusLine statusLine = closeableHttpResponse.getStatusLine();

        String content;
        try {
            HttpEntity httpEntity = closeableHttpResponse.getEntity();
            content = IOUtils.toString(httpEntity.getContent(), Charset.defaultCharset());
            EntityUtils.consume(httpEntity);
        } finally {
            closeableHttpResponse.close();
        }

        if (closeableHttpResponse.getStatusLine().getStatusCode() >= 400) {
            LOGGER.error("Request failed with {}, response body is: \n{}", statusLine, content);
        }

        return content;
    }
}
